import numpy as np

def choice(a, size=None, replace=True):
    m = len(a)
    # Single sample
    if size==None:
        random_i = np.random.randint(m)
        return a[random_i]
    else:
        num_sample = np.prod(size)

    # Multi-Sample
    a=np.asarray(a)
    if replace==True:
        random_i = np.random.randint(m, size = num_sample)
        return a[random_i].reshape(size)
    else:
        assert m >= num_sample, 'entries of array cannot exceed number of samples'
        random_i = np.arange(m)
        np.random.shuffle(random_i)
        random_i = random_i[:num_sample]
        return a[random_i].reshape(size)



# %%
import torch
import torch.nn as nn

# %%
def aggregate(model_source, model_target, weight = None):
    '''
    model_source: List of nn.Model instances
    model_target: Single nn.Model instance
    weights: default = None
        List of numbers. Must match the number of model_source.
        If None, then weight is set as 1/len(model_source)
    '''
    if weight == None:
        weight = [1/len(model_source)] * len(model_source)
    assert len(model_source) == len(weight), "length of model_source(%s) and weight(%s) does not match"%(len(model_source), len(weight))

    for parameters in zip(model_target.parameters(), *[model.parameters() for model in model_source]):
        p_trg = parameters[0]
        p_src_tuple = parameters[1:]
        # model_target's device
        device = p_trg.device

        # 1. Reset
        p_trg.data[:] = 0

        # 2. Add weighted sum
        for p_src, w in zip(p_src_tuple, weight):
            p_trg.data += (w * p_src.data).to(device)

def distribute(model_source, model_target):
    '''
    model_source: Single nn.Model instance
    model_target: List of nn.Model instances
    '''
    for parameters in zip(model_source.parameters(), *[model.parameters() for model in model_target]):
        p_src = parameters[0]
        p_trg_tuple = parameters[1:]

        for p_trg in p_trg_tuple:
            device = p_trg.device
            p_trg.data[:] = p_src.data.to(device)

def aggregate_grad(model_source, model_target):
    '''
    model_source: List of nn.Model instances
    model_target: Single nn.Model instance
    '''
    for parameters in zip(model_target.parameters(), *[model.parameters() for model in model_source]):
        p_trg = parameters[0]
        p_src_tuple = parameters[1:]
        # model_target's device
        device = p_trg.device

        for p_src in p_src_tuple:
            p_trg.grad += p_src.grad.to(device)

def distribute_all(model_source, model_target):
    '''
    model_source: Single nn.Model instance
    model_target: List of nn.Model instances
    '''
    for model_target_ in model_target:
        model_target_.load_state_dict(model_source.state_dict())

def aggregate_state_dict(model_list, device):
    '''
    aggregate model values which are not parameters(weight&bias) but are updated
    '''
    state_dict = {}
    state_dict_list = []
    for model in model_list:
        print(next(model.parameters()).device)
        for key, value in model.state_dict().items():
            if 'weight' in key or 'bias' in key:
                continue
            if key not in state_dict:
                state_dict[key] = value.clone().to(device)
            else:
                state_dict[key] += value.to(device)

    n_model = len(model_list)
    for key in state_dict:
        state_dict[key] /= n_model

    return state_dict

def attentive_aggregate(model_source_list, model_target, step_size = 0.01, p = 2, optimizer = None):
    '''
    perform attentive aggregation from source -> target.
    model_source_list: List of nn.Model instances
    model_target: Single nn.Model instance
    '''
    for parameters in zip(model_target.parameters(), *[model.parameters() for model in model_source]):
        p_trg = parameters[0]
        p_src_tuple = parameters[1:]
        # model_target's device
        device = p_trg.device

        # 1. s = | w_server - w_cleint |
        delta = []
        for p_src in p_src_tuple:
            delta.append(p_trg - p_src.to(device))
        delta = torch.as_tensor(delta).to(device)
        assert len(delta) == len(p_src_tuple)

        s_k = torch.norm(delta, p=p, dim = tuple(range(1, len(p_trg.shape)+1 ) ) ) # dim = (1,2) or (1,2,3) or (1,2,3,...)
        assert len(s_k) == len(p_src_tuple)

        # 2. attention = softmax(s, dim = client)
        attention = torch.softmax(s_k, dim = 0)

        # 3. store gradient
        gradient = torch.sum(attention.expand(delta.shape[::-1]).T * delta, dim = 0)
        p_trg.grad[:] = gradient

        # 4. apply gradient
        p_trg.data -= step_size * p_trg.grad
        # (possible to use optimizer here?)


        # # 1. s = | w_server - w_cleint |
        # s_k = []
        # for p_src in p_src_tuple:
        #     s_k.append(torch.norm(p_trg - p_src.to(device), p=p)) # matrix norm of the difference of layer weights
        # s_k = torch.as_tensor(s_k).to(device)
        # assert len(s_k) == len(p_src_tuple)
        #
        # # 2. attention = softmax(s, dim = client)
        # attention = torch.softmax(s_k, dim = 0)
        #
        # # 3. store gradient
        # p_trg
        # gradient = []
        # for a, p_src in zip(attention, p_src_tuple):
        #     gradient.append(a * (p_trg - p_src.to(device) )


        # 4. apply gradient (either by adding or )

# from model import DNN
# import torch
# import torch.nn as nn
# import torch.nn.functional as F
# import torch.optim as optim
# n_input = 5
# n_output = 10
# n_hidden_list = [10,20,30,20]
# n_model = 4
# activation = [nn.Sigmoid(), nn.ReLU(), nn.LeakyReLU(0.03), nn.Tanh(), nn.Identity()]
# model_list = [DNN(n_input, n_output, n_hidden_list, activation) for i in range(n_model)]
#
# model_source = model_list[1:]
# model_target = model_list[0]
# model_target.cuda()
#
# x = torch.randn(20,5)
# for model in model_list:
#     device = next(model.parameters()).device
#     y_hat = model(x.to(device))
#     loss = torch.sum(y_hat)
#     loss.backward()
#
# p=list(model_target.parameters())
# p[0].grad
# next(model_source[0].parameters()).grad
#
# g = torch.zeros(10,5)
# for model in model_list:
#     g += next(model.parameters()).grad.cpu()
# print('cumsum of gradient:\n',g)


# %%
# class m(nn.Module):
#     def __init__(self):
#         super(m, self).__init__()
#         self.x = nn.Linear(10,10)
# # %%
# model_source = [m().cuda() for i in range(5)]
# model_target = m()
# for i, model in enumerate(model_source):
#     model.x.weight.data[:] = i
# # %%
# for model in model_source:
#     print(model.x.weight)
# print(model_target.x.weight.data)
# # %%
# aggregation(model_source, model_target)
# # %%
# print(model_target.x.weight.data)
# # %%
# distribution(model_target, model_source)
# # %%
#
# y.x.weight.data
# weight = None
# for i in zip([t.parameters() for t in model_source]):
#     print(i)
# torch.cuda.device_count()



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

class DNN(nn.Module):
    '''
    basic DNN module

    Parameters
    ----------
    n_input: number of input (int)
    n_output: number of output (int)
    n_hidden_list: list of hidden neurons (list of int)
    activation: torch.nn activation function instances (nn activation instance or list)
    '''
    def __init__(self, n_input, n_output, n_hidden_list, activation):
        super(DNN, self).__init__()
        if type(activation) is not list:
            activation = [activation]*(len(n_hidden_list)+1)
        assert len(activation)==len(n_hidden_list)+1, 'length of layers and activations must match. If you want no activation, use nn.Identity'

        layers = [nn.Linear(n_input, n_hidden_list[0]), activation[0]]
        for i in range(len(n_hidden_list) - 1):
            layers.extend([nn.Linear(n_hidden_list[i], n_hidden_list[i+1]), activation[i+1]])
        layers.extend([nn.Linear(n_hidden_list[-1], n_output), activation[-1]])

        self.fc = nn.Sequential(*layers)

    def forward(self, x):
        return self.fc(x)

if __name__ == '__main__':
    n_input = 5
    n_output = 10
    n_hidden_list = [10,20,30,20]
    activation = [nn.Sigmoid(), nn.ReLU(), nn.LeakyReLU(0.03), nn.Tanh(), nn.Identity()]
    model = DNN(n_input, n_output, n_hidden_list, activation)
    print(model)



import pickle
import time
import subprocess

def save_pickle(obj, path = None):
    if path == None:
        path = time.strftime('%y%m%d_%H%M%S.p')
    with open(path, 'wb') as f:
        pickle.dump(obj, f)

def load_pickle(path):
    with open(path, 'rb') as f:
        return pickle.load(f)

def write(content, path, encoding = None):
    with open(path, 'w', encoding = encoding) as f:
        f.write(content)

def append(content, path, encoding = None):
    with open(path, 'a', encoding = encoding) as f:
        f.write(content)

def read(path, encoding = None):
    with open(path, 'r', encoding = encoding) as f:
        text = f.read()
    return text

def readline(path, encoding = None):
    '''Create generator object which iterates over the file'''
    with open(path, 'r', encoding = encoding) as f:
        for line in f:
            yield line

def readlines(path, encoding = None):
    with open(path, 'r', encoding = encoding) as f:
        text = f.readlines()
    return text

def cmd(command):
    pipe = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, encoding='utf-8')
    return pipe.stdout

# Used in Argparse
def str2bool(x):
    true_list = ['t', 'true', 'y', 'yes', '1']
    false_list = ['f', 'false', 'n', 'no', '0']
    if x.lower() in true_list:
        return True
    elif x.lower() in false_list:
        return False
    else:
        raise Exception('input has to be in one of two forms:\nTrue: %s\nFalse: %s'%(true_list, false_list))

def is_iterable(x):
    return hasattr(x, '__iter__')

class tdict(dict):
    def __init__(self, *args, **kwargs):
        super(tdict, self).__init__(*args, **kwargs)

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __delattr__(self, key):
        self.__delitem__(key)

class Printer():
    def __init__(self, filewrite_dir = None):
        self.content = ''
        self.filewrite_dir = filewrite_dir

    def add(self, text):
        self.content += text

    def print(self, *args, end='\n', flush=False):
        self.add(' '.join([str(arg) for arg in args]))
        print(self.content, end=end, flush=flush)
        if self.filewrite_dir is not None:
            append(self.content + end, self.filewrite_dir)
        self.content=''

    def reset(self):
        self.content = ''

class Timer():
    '''Timer to measure elapsed time

    Parameters
    ----------
    print: bool (default: False)
        if True, then prints self.elapsed_time whenever stop() is called.

    return_f: bool (default: True)
        if True, then returns self.elapsed_time whenever stop() is called.

    auto_reset: bool (default: True)
        if True, then resets whenever start() is called.

    Methods
    -------
    start

    stop

    reset
        sets self.elapsed_time = 0
    '''
    def __init__(self, print = False, return_f = True, auto_reset = True):
        self.print = print
        self.return_f = return_f
        self.auto_reset = auto_reset
        self.elapsed_time = 0
        self.running = False

    def __enter__(self):
        self.start()

    def __exit__(self):
        t = self.stop()
        print(t)

    def start(self):
        if self.auto_reset:
            # self.reset()
            self.elapsed_time = 0
        self.running = True
        self.start_time = time.time()

    def stop(self):
        if self.running:
            self.end_time = time.time()
            self.elapsed_time += self.end_time - self.start_time
            self.running = False
        if self.print:
            print(self.elapsed_time)
        if self.return_f:
            return self.elapsed_time

    def reset(self):
        self.elapsed_time = 0

class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count



from torch.optim.lr_scheduler import _LRScheduler
class AdaptiveLR(_LRScheduler):
    """Modify the learning rate of each parameter group,
    depending on the current & previous loss values.

    If current_loss < previous_loss:
        lr += a*lr (Increment lr by a * lr)
    If current loss > previous loss:
        lr -= b*lr (Decrement lr by b * lr)

    Args:
        optimizer (Optimizer): Wrapped optimizer.
        a: (float or list): A multiplicative factor to increment learning rate
            when the current loss is lower than the previous loss.
        b: (float or list): A multiplicative factor to decrement learning rate
            when the current loss is higher than the previous loss.

        The increment/decrement of learning rate is done on
        each lr for each group in optimizer.param_groups
    Example:
        >>> # Assuming optimizer has two groups.
        >>> a_list = [0.1, 0.2]
        >>> b_list = [0.5, 0.9]
        >>> scheduler = AdaptiveLR(optimizer, a=a_list, b=b_list)
        >>> for epoch in range(100):
        >>>     train(...)
        >>>     validate(...)
        >>>     scheduler.step(loss)
    """
    requires_loss = True
    def __init__(self, optimizer, a, b, last_epoch = -1, threshold = -float('inf'), fix = False, last_loss = float('inf')):
        self.last_loss = last_loss
        self.loss = self.last_loss
        self.threshold = threshold
        self.fix = fix
        self.factor = 1
        self.a = a
        self.b = b
        # if (not isinstance(a, list) and not isinstance(a, tuple)) and (not isinstance(b, list) and not isinstance(b, tuple)):
        #     self.a_s = [a] * len(optimizer.param_groups)
        #     self.b_s = [b] * len(optimizer.param_groups)
        # elif (isinstance(a, list) or isinstance(a, tuple)) and (isinstance(b, list) or isinstance(b, tuple)):
        #     if len(a) != len(optimizer.param_groups) or len(b) != len(optimizer.param_groups):
        #         raise ValueError("Expected {} 'a's and 'b's, but got {}, {}".format(
        #             len(optimizer.param_groups), len(a), len(b)))
        #     self.a_s = list(a)
        #     self.b_s = list(b)
        # else:
        #     raise ValueError("a and b has to be the same type, of either iterable or numeric values."
        #                     "Got: a: {}, b: {}".format(type(a), type(b)))
        super(AdaptiveLR, self).__init__(optimizer, last_epoch)

    def step(self, loss = 0, epoch=None):
        if self.last_epoch >= 0:
            self.last_loss = self.loss
            self.loss = loss
        # self.last_epoch gets incremented afterwards
        super(AdaptiveLR, self).step(epoch=epoch)

    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn("To get the last learning rate computed by the scheduler, "
                          "please use `get_last_lr()`.", DeprecationWarning)
        if self.last_epoch > 0:
            if self.factor < self.threshold:
                # lr fixed after threshold
                if self.fix:
                    return [base_lr * self.threshold for base_lr in self.base_lrs]
                # lr not fixed after threshold
                elif self.loss <= self.last_loss:
                    self.factor += self.factor * self.a
                    return [base_lr * self.factor for base_lr in self.base_lrs]
                    # return [group['lr'] + group['lr'] * self.a for group in self.optimizer.param_groups]
                else:
                    return [base_lr * self.factor for base_lr in self.base_lrs]
                    # return [group['lr'] for group in self.optimizer.param_groups]
            else:
                if self.loss <= self.last_loss:
                    self.factor += self.factor * self.a
                    return [base_lr * self.factor for base_lr in self.base_lrs]
                    # return [group['lr'] + group['lr'] * self.a for group in self.optimizer.param_groups]
                else:
                    self.factor -= self.factor * self.b
                    return [base_lr * self.factor for base_lr in self.base_lrs]
                    # return [group['lr'] - group['lr'] * self.b for group in self.optimizer.param_groups]
        else:
            return self.base_lrs

class LinearLR(_LRScheduler):
    """Decay the learning rate of each parameter group by delta every epoch.
    Every step,
    lr = lr - initial_lr * delta
    -----
    Args:
        optimizer (Optimizer): Wrapped optimizer.
        delta: (float or list): A subtractive factor to decrement learning rate.
        last_epoch (int): The index of last epoch. Default: -1.
    Example:
        >>> # Assuming optimizer has two groups.
        >>> delta_list = [0.1, 0.2]
        >>> scheduler = AdaptiveLR(optimizer, delta=delta_list)
        >>> for epoch in range(100):
        >>>     train(...)
        >>>     validate(...)
        >>>     scheduler.step()
    """
    def __init__(self, optimizer, delta, last_epoch = -1):
        self.deltas = delta
        if not isinstance(delta, list) and not isinstance(delta, tuple):
            self.deltas = [delta] * len(optimizer.param_groups)
        else:
            if len(delta) != len(optimizer.param_groups):
                raise ValueError("Expected {} deltas, but got {}".format(
                    len(optimizer.param_groups), len(delta)))
            self.deltas = list(delta)
        super(LinearLR, self).__init__(optimizer, last_epoch)

    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn("To get the last learning rate computed by the scheduler, "
                          "please use `get_last_lr()`.", DeprecationWarning)
        if self.last_epoch > 0:
            return [group['lr'] - base_lr * delta for group, base_lr, delta in zip(self.optimizer.param_groups, self.base_lrs, self.deltas)]
        else:
            return self.base_lrs

    def _get_closed_form_lr(self):
        return [base_lr*(1 - delta * self.epoch) for base_lr, delta in zip(self.base_lrs, self.deltas)]


class NoneLR(_LRScheduler):
    def __init__(self, optimizer, last_epoch = -1):
        super(NoneLR, self).__init__(optimizer, last_epoch)
    def get_lr(self):
        return self.base_lrs
#
# import torch
# import torch.optim as optim
# w = torch.randn((1,5), requires_grad=True)
# y = torch.randn((2,3), requires_grad=True)
# optimizer = optim.Adam([w,y])
# optimizer
# lr_schduler = AdaptiveLR(optimizer, a=[0.1,0.2], b=0.5)
# help(AdaptiveLR1)
# lr_schduler.state_dict()
#
# for
# optimizer.step()
# lr_schduler.step()
# lr_schduler.state_dict()
#
# class AdaptiveLR():
#     def __init__(self, optimizer, a, b, last_epoch = 0):
#         self.optimizer = optimizer
#         self.a = a
#         self.b = b
#         self.last_loss = float('inf')
#         self.base_lrs = [param_group['lr'] for param_group in optimizer.param_groups]
#         self.last_epoch = last_epoch
#         self._last_lr = self.base_lrs
#
#     def state_dict(self):
#         content = dict(
#         optimizer = self.optimizer,
#         a = self.a,
#         b = self.b,
#         last_loss = self.last_loss,
#         base_lrs = self.base_lrs,
#         last_epoch = self.last_epoch,
#         _last_lr = self._last_lr,
#         )
#         return content
#
#     def step(self, loss):
#         for i, param_group in enumerate(self.optimizer.param_groups):
#             lr = param_group['lr']
#             if loss <= self.last_loss:
#                 self._last_lr[i] += lr * self.a
#             else:
#                 self._last_lr[i] -= lr * self.b
#             self.optimizer.param_groups[i]['lr'] = self._last_lr[i]
#         self.last_loss = loss
#         self.last_epoch += 1
#
#
#     def get_last_lr(self):
#         return self._last_lr

def lr_schedule_VAE(epoch):
    if epoch <= 100:
        lr_target = 1e-4
        lr_decay = (lr_target) ** (1/100 * epoch)
        return lr_decay
    else:
        return 1e-3
